# version 1.21, changed by DD, 2010.03.08
# version 1.2, changed by LB, 20100303_FR21:19

LIFT_Fr := function(machine,image,addresses)
    # lift <machine> by choosing the starting points at <addresses>, which
    # is a list of integers in [1..degree].
    # <image> is such that recursion of <image>[i] contains generator i.
    # returns the isomorphism of the stateset that
    # takes from the old to the new machine.
    local tau, gens, g, h, j, s, c, a;
    
    gens := GeneratorsOfFRMachine(machine);
    tau := [];
    for s in [1..Length(gens)] do
        g := gens[image[s]];
        a := addresses[s];
        h := One(gens[1]);
        repeat
            h := h*Transition(machine,g,a);
            a := Output(machine,g,a);
        until a=addresses[s];
        if not IsConjugate(StateSet(machine),gens[s],h) then
            return fail;
        fi;
        Add(tau,h);
    od;
    return GroupHomomorphismByImages(StateSet(machine),StateSet(machine),tau,gens);
end;

LIFTONE_Fr := function(machine,image,start,j)
    # lift <machine> by choosing the starting points at <addresses>, which
    # is a list of integers in [1..degree].
    # <image> is such that recursion of <image>[i] contains generator i.
    # returns the isomorphism of the stateset that
    # takes from the old to the new machine.
    local tau, gens, g, h, s, c, a;

    gens := GeneratorsOfFRMachine(machine);
    tau := ShallowCopy(gens);
    g := gens[image[j]];
    a := start;
    h := One(gens[1]);
    #Print(start," ",j," ",g,"\n");
    repeat
        h := h*Transition(machine,g,a);
        a := Output(machine,g,a);
        #Print(h,"  ",a,"\n");
    until a=start;
    #Print(start,j,h,"\n\n");
    if not IsConjugate(StateSet(machine),gens[j],h) then
        return fail;
    fi;
    tau[j] := h;
    return GroupHomomorphismByImages(StateSet(machine),StateSet(machine),tau,gens);
end;


ADMADDRESSES_Fr := function(machine)
    # returns a pair [image,[v1...vm]], with
    # vi = the coordinates in <machine> of a cycle whose product is
    # conjugate to vi, and
    # image[i] = the index of the element with such a cycle.
    # return [fail,fail] if these cannot be found, because e.g. there are
    # two or no cycles whose product is conjugate to the same vi.
    
    local gens, g, j, s, c, image, addresses;
    
    image := [];
    addresses := [];
    gens := GeneratorsOfFRMachine(machine);
    for s in [1..Length(gens)] do
        for c in Cycles(PermList(Output(machine,gens[s])),AlphabetOfFRObject(machine)) do
            g := One(gens[1]);
            for j in c do
                g := g*Transition(machine,gens[s],j);
            od;
            if IsOne(g) then continue; fi;
            g := CyclicallyReducedWord(g);
            g := Position(gens,g);
            if g=fail or IsBound(image[g]) then return [fail,fail]; fi;
            image[g] := s;
            addresses[g] := c;
        od;
    od;
    if BoundPositions(image)<>[1..Length(gens)] then
        return [fail,fail];
    fi;
    return [image,addresses];
end;

COUNTBOOL_Fr := function(w) # return fail if one of the states of a recursion w contains at least 2 equal letters
    local v,i,count;
    i:=1;
    v := ExtRepOfObj(w);
    count := [];
    for i in [2,4..Length(v)] do
        if not IsBound(count[v[i-1]]) then count[v[i-1]] := 0; fi;
        count[v[i-1]] := count[v[i-1]] + AbsInt(v[i]);
        if count[v[i-1]]>2 then return false; fi;
    od;
    return true;
end;

LEVYCYCLES_Fr := function(machine,image) #return binary relation on the postcritical set
    # 
    local tau, gens, g, h, s, set, sset, c, a, v,i,j,k,R;
   
    gens := GeneratorsOfFRMachine(machine);    
    set:=[1..Length(gens)];
    sset:=[];
   #R:= EquivalenceRelationByPartition(Domain([1..Length(gens)]), [[1..Length(gens)]]);

   # image := [];
   # addresses := [];

    for s in [1..Length(gens)] do
        for c in Cycles(PermList(Output(machine,gens[s])),AlphabetOfFRObject(machine)) do
    #        g := One(gens[1]);
            if Length(c)>1 then
               for j in [c[1]..c[Length(c)-1]] do
                   v:=ExtRepOfObj(Transition(machine,gens[s],j));
                   for k in [1..Length(v)/2] do
                      Remove(v,k+1); 
                   od;
                   Add(sset,v);
                   set:=Difference(set,v);
               od;
            fi;
   #        if IsOne(g) then continue; fi;
   #        g := CyclicallyReducedWord(g);
   #        g := Position(gens,g);
   #        if g=fail or IsBound(image[g]) then return [fail,fail]; fi;
   #         image[g] := s;
   #         addresses[g] := c;
        od;
    od;
#Print("\n",set,sset);
   Add(sset,set);
   R:= EquivalenceRelationByPartition(Domain([1..Length(gens)]), sset); 




   # tau := ShallowCopy(gens);
   # g := gens[image[j]];
   # a := start;
   # h := One(gens[1]);
    #Print(start," ",j," ",g,"\n");
   # repeat
   #     h := h*Transition(machine,g,a);
   #     a := Output(machine,g,a);
   #     #Print(h,"  ",a,"\n");
   # until a;#a=start;
   # i:=1;
    #v := ExtRepOfObj(w);
   # count := [];
   # for i in [2,4..Length(v)] do
   #     if not IsBound(count[v[i-1]]) then count[v[i-1]] := 0; fi;
   #     count[v[i-1]] := count[v[i-1]] + AbsInt(v[i]);
   #     if count[v[i-1]]>2 then return false; fi;
   # od;
    return R;
end;




SPIDERALGORITHM_Fr := function(machine)
    # <machine> is a polynomial FR machine:
    # StateSet(M) is a free group of rank m+1;
    # generator AddingElement(M) is an adding element, and should be ignored.
    
    local image, # which point maps to which one
          numcycles, # number of cycles in "image"
          preimage, # either inverse of image, for periodic points, or 0
          preperiodicity, # distance to periodic cycle
          gens, # generators of free group
          niter, # number of iterations
          m, # number of generators
          degree, #degree   
          suppangles,
          hyperbolic, # whether cycle contains a critical point
          cyclelength, # length of cycle
          orbit, # cycle index of point
          marked, # one marked point per grand cycle
          pass, # 1,2,3 for each cycle
          wobble, # little extra push to give to some markings
          oldmarkrecursion, # previous recursions of marked elements
          tau, tauproduct, # the group automorphisms
          group, # the free group
	  mainmarktau, # sequence of \tau_1,tau_1*tau_2,...
	  marktau, # words describing the ray to the previous markings
 	  rayperiodindex, # a parameter for every subhyperbolic cycle 
          coincidence,
          address, # where we lift our machine
          markedaddress, # orbit of marked points
          admaddresses, # admissible addresses for lifting
          rayperiod, # period of the supporting rays landing there
          i, j, k, p, t,machine2;
    
    # initialise constants

    degree := Length(AlphabetOfFRObject(machine));
    gens := GeneratorsOfFRMachine(machine);
    group := StateSet(machine);
    hyperbolic := ADMADDRESSES_Fr(machine);
    image := hyperbolic[1];
    if image=fail then return -1; fi;
    m := Length(gens);
    orbit := image;
    for i in [1..m] do orbit := image{orbit}; od;
    cyclelength := List(orbit,x->[x]);
    for i in [1..m] do
        cyclelength := Set(cyclelength,o->Union(o,image{o}));
    od;
    numcycles := Size(cyclelength);
    j := Union(cyclelength);
    preimage := List([1..m],i->0); preimage{image{j}} := j;
    preperiodicity := [];
    for i in [1..m] do
        niter := 0;
        k := i; while not k in j do k := image[k]; niter := niter+1; od;
        preperiodicity[i] := niter;
    od;
    hyperbolic := List(cyclelength,o->ForAny(o,i->Length(hyperbolic[2][i])>1));
    orbit := List(orbit,i->First(cyclelength,c->i in c));
    marked := List(cyclelength,Representative);
    cyclelength := List(cyclelength,Length);

    # initialise variables
    niter := 0;
    # marked was already initialised above
    pass := List([1..numcycles],i->1);
    wobble := List([1..numcycles],i->0);
    oldmarkrecursion := List([1..numcycles],i->[WreathRecursion(machine)([marked[i]])]);
    tau := IdentityMapping(group);
    tauproduct := IdentityMapping(group);  #do we need it???
    mainmarktau :=[];#List([1..m],x->IdentityMapping(group));#[];#do we need it???
    marktau := List([1..numcycles],i->List([1..m],j->gens[1]));#do we need it???
    coincidence := List([1..numcycles],i->0);
    rayperiodindex:=List([1..numcycles],i->0);
    markedaddress := [];

    repeat
        # step 1: compute possible addresses
        admaddresses := ADMADDRESSES_Fr(machine)[2];
        if admaddresses=fail then return -1; fi;
        if niter=0 then
            address := List(admaddresses,Minimum);
        fi;
Display(machine);
        # step 2: select appropriate lifts
        for i in [1..numcycles] do if hyperbolic[i] then
            j := marked[i];
            if niter < cyclelength[i] then
                address[j] := Minimum(admaddresses[j]);
            else
                k:= Intersection(admaddresses[j],[markedaddress[j]+wobble[i]..m]);  
                if k=[] then
                    address[j] := Minimum(admaddresses[j]);
                    wobble[i] := 1;
                else
                    address[j]:=Minimum(k); 
                    wobble[i] := 0;
                fi;
            fi;
            markedaddress[j] := address[j];
            marked[i] := preimage[j];
        fi; od;


        # step 3: check addresses
        for j in [1..m] do
            if not address[j] in admaddresses[j] then
                #!!!! same change as above???
                k := Intersection(admaddresses[j],[address[j]+1..m]);
                if k=[] then
                    address[j] := Minimum(admaddresses[j]);
                else
                    address[j]:=Minimum(k);
                fi;
            fi;
        od;
        #Info(InfoFR, 2, "Symbolic spider: selected addresses ", address);


        # step 4: do the lifting
        tau := LIFT_Fr(machine,image,address);

        #Error("test",tau); #!!! should be tau^-1 somewhere # !!! type return to continue
        if tau=fail then return -2; fi;
#???        Print(Successors(LEVYCYCLES_Fr(machine*tau,image)),"\n");#???
        Display(machine); 
        Print(tau);
        p:=machine*tau;
        Display(p);
        machine:=machine^tau; 
#???
#???    machine:=tau^(-1)*(machine*tau);  
        
        for i in [1..numcycles] do
            if not (oldmarkrecursion[i] = false) then #???
                #oldmarkrecursion[i][1] := List(oldmarkrecursion[i][1],x->x^tau); #???
                 if ForAll([1..Length(oldmarkrecursion[i])],
                         t-> ForAll( [1..Length(oldmarkrecursion[i][t])], k-> COUNTBOOL_Fr(oldmarkrecursion[i][t][1][k]) ) ) 
                  then 
                     for j in [1..Length(oldmarkrecursion[i])] do
                        oldmarkrecursion[i][j][1] :=  List(oldmarkrecursion[i][j][1],x->x^tau);                  
                     od;
                 else 
                     oldmarkrecursion[i] := false;
                 fi;  
	 fi;
        od;
 
        niter := niter + 1;
   #     if niter>=m+1 then Remove(mainmarktau); fi; #???

   #    Info(InfoFR, 2, "Symbolic spider: new machine ", machine); #???
        # step 5: adjust pass
        for i in [1..numcycles] do
            if pass[i]=3 then continue; fi;
            if hyperbolic[i] then
                if niter mod cyclelength[i] <> 0 then continue; fi;
                j := WreathRecursion(machine)(gens[marked[i]]);
                if oldmarkrecursion[i] = false then oldmarkrecursion[i] := [j]; continue; fi; 
                if j=oldmarkrecursion[i][1] then
                    pass[i] := pass[i]+1;
                   # if pass[i]=2 then
                   #   oldmarkrecursion[i][1] := false;
                   #   wobble[i]:=1-wobble[i];
                   # fi;
                    wobble[i] := 3-pass[i]; # 1 on pass 2, 0 on pass 1&3
                else
                    oldmarkrecursion[i][1] := j;
                fi;
            else   #Almost the same???
                if niter mod cyclelength[i] <> 0 then continue; fi;
                j := WreathRecursion(machine)(gens[marked[i]]);  
                if oldmarkrecursion[i] = false then oldmarkrecursion[i] := [j]; continue; fi; 
                if j in oldmarkrecursion[i] then
                    pass[i] := 3;
  #     Rayperiod???        
                else
                    Add(oldmarkrecursion[i],j,1); 
                    if Length(oldmarkrecursion[i]) > m then Remove(oldmarkrecursion[i]); fi;
                fi;
            fi;
        od;

#Print(niter,"   ",markedaddress," \n",wobble," ",address,pass,"\n");   
    until ForAll([1..numcycles],i-> pass[i]=3); 
              #or  (not hyperbolic[i] and coincidence[i]>=m));  ???
#for i in [1..Rank(StateSet(machine))] do
#    Print(WreathRecursion(machine)(gens[i]),"\n");
#od

Display(machine);

# step 7: adjust the marked addresses
   for i in [1..numcycles] do
        if hyperbolic[i] then
            j := preimage[marked[i]];#??? preimage[marked[i]]
            while j<>marked[i] do
                tau := LIFTONE_Fr(machine,image,markedaddress[j],j);#???markedaddress[image[j]]
                if tau=fail then return -3; fi;
                if IsBijective(tau)=false then return -3; fi;
                machine := machine^tau;
                #tauproduct := tauproduct*tau;
                j := preimage[j];
            od;
        fi;
        if not (preperiodicity{Filtered([1..m],j->orbit[j]=i)}=[]) then
            for k in [1..Maximum(preperiodicity{Filtered([1..m],j->orbit[j]=i)})] do
                for j in [1..m] do if orbit[j]=i and preperiodicity[j]=k then
                   admaddresses := ADMADDRESSES_Fr(machine)[2];
                   markedaddress[j] := Minimum(admaddresses[j]);
                   tau := LIFTONE_Fr(machine,image,markedaddress[image[j]],j);
                   if tau=fail then return -3; fi;
                   if IsBijective(tau)=false then return -4; fi;
                   machine := machine^tau;
                   #tauproduct := tauproduct*tau;
                 fi; od;
            od;
        fi;
    od;

#for i in [1..Rank(StateSet(machine))] do
#    Print(WreathRecursion(machine)(gens[i]),"\n");
#od;

suppangles:=[]; # Calculating supporting rays

for j in [1..m] do
   if Length(admaddresses[j])=1 then continue; fi; 
   p:=markedaddress[j]-1;
   k:=j;
   t:=degree;
   i:=[];
   Add(i,k);
   k:=image[k];
   while not k in i do
      t:=t*degree;
      p:=p*degree+markedaddress[k]-1;#p:=List([1..Length(p)],i->p[i]*degree+markedaddress[k]-1);
      Add(i,k); 
      k:=image[k];
   od; 
   p:=p/(t-1);
   Add(suppangles,List([1..Length(admaddresses[j])],i->(admaddresses[j][i]-markedaddress[j])/degree+p));
 od;
#Print(niter);
    return rec(suppangles :=suppangles);#,
             #  cleanmachine := machine,
             #  degree := Length(AlphabetOfFRObject(machine)),
             #  image := image,
             #  markedaddress := markedaddress);
end;

# machine:=PolynomialIMGMachine(12,[[1/(12^12-1),1/(12^12-1)+1/12,1/(12^12-1)+2/12,1/(12^12-1)+3/12,1/(12^12-1)+4/12,1/(12^12-1)+5/12,1/(12^12-1)+6/12,1/(12^12-1)+7/12,1/(12^12-1)+8/12,1/(12^12-1)+9/12,1/(12^12-1)+10/12,1/(12^12-1)+11/12]],[]);
# group:=StateSet(machine);
# T:=GroupHomomorphismByImages(group,group,[group.1^(group.2*group.1),group.2^group.1,group.3,group.4,group.5^(group.6*group.5),group.6^group.5,group.7,group.8,group.9,group.10,group.11,group.12,group.13],GeneratorsOfGroup(group));


